from abc import ABCMeta, abstractmethod
from collections import deque
from contextlib import suppress
from functools import wraps
from asyncio import (
        get_event_loop, sleep, CancelledError,
        run_coroutine_threadsafe
    )

from .waiter import Waiter, EmptyWaiter
from .interface import asyncify


class ClosedChannel(Exception):
    """Channel is closed, can't send any more messages."""

    def __init__(self):
        super().__init__("Channel is closed, can't send any more messages.")


def _common_chan_call(call):
    """Code of __call__ common to both types of channels."""
    @wraps(call)
    async def method(self, *messages, **kwargs):
        if self.closed:
            raise ClosedChannel

        if not messages:
            message = None
        elif len(messages) == 1:
            message = messages[0]
        else:
            message = tuple(messages)

        return await call(self, message, **kwargs)

    return method


class _AutoThreadify(ABCMeta):
    """Make thread-safe versions of channel's methods.

    It's a metaclass so that channels can be easily inherited.
    """

    methods = {"__call__": "send", "close": "close_threadsafe",
               "get": "recv", "close_nowait": "close_nowait_threadsafe"}

    def __new__(*args, **kwargs):
        cls = type.__new__(*args, **kwargs)

        for orig_name, ts_name in _AutoThreadify.methods.items():
            original = getattr(cls, orig_name)
            if orig_name == "close_nowait":
                original = asyncify(original)

            threadsafe = _AutoThreadify.threadify(original)
            setattr(cls, ts_name, threadsafe)

        return cls

    def threadify(method):
        @wraps(method)
        def threadified(self, *args, **kwargs):
            coroutine = method(self, *args, **kwargs)
            future = run_coroutine_threadsafe(coroutine, self.loop)
            return future.result()

        return threadified


class _ChannelCommon(metaclass=_AutoThreadify):

    __slots__ = "loop", "_input", "_output", "closed"

    def __init__(self, *, loop=None, input_max=None, output_max=None):
        self.loop = loop or get_event_loop()
        self._output = Waiter(loop=self.loop, maxsize=output_max)
        self._input = Waiter(loop=self.loop, maxsize=input_max)
        self.closed = False

    @abstractmethod
    def get(self):
        pass  # pragma: no cover

    @abstractmethod
    async def __call__(self, message):
        pass  # pragma: no cover

    @abstractmethod
    async def close(self):
        pass  # pragma: no cover

    def close_nowait(self):
        if self.closed:
            return
        self.closed = True
        closed_exception = ClosedChannel()
        self._output.exception(closed_exception)
        self._input.exception(closed_exception)

    def __await__(self):
        return self.get().__await__()

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return await self
        except (ClosedChannel, CancelledError):
            raise StopAsyncIteration

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.recv()
        except (ClosedChannel, CancelledError):
            raise StopIteration

    @abstractmethod
    def recv(self):
        pass  # pragma: no cover

    @abstractmethod
    def send(self, *messages):
        pass  # pragma: no cover

    @abstractmethod
    def close_threadsafe(self):
        pass  # pragma: no cover

    @abstractmethod
    def close_nowait_threadsafe(self):
        pass  # pragma: no cover


class Channel(_ChannelCommon):

    async def get(self):
        if self.closed:
            raise ClosedChannel

        with suppress(EmptyWaiter):
            self._input(True)

        return await self._output

    @_common_chan_call
    async def __call__(self, message):
        try:
            self._output(message)
        except EmptyWaiter:
            await self._input
            # Yield control back to __await__ so it can put something
            # into _input before we try to send a result to it.
            await sleep(0)
            self._output(message)

    async def close(self):
        self.close_nowait()


class BufferedChannel(_ChannelCommon):

    __slots__ = "maxsize", "_buffer", "_closing"

    def __init__(self, **kwargs):
        maxsize = None
        with suppress(KeyError):
            maxsize = kwargs["maxsize"]
            del(kwargs["maxsize"])
        super().__init__(**kwargs)
        self._buffer = deque(maxlen=maxsize)
        self._closing = None
        self.maxsize = maxsize

    async def get(self):
        if self.closed:
            raise ClosedChannel
        if self.empty:
            await self._output

        message = self._buffer.popleft()

        with suppress(EmptyWaiter):
            self._input(None)

        if self._closing and len(self._buffer) == 0:
            self._closing.set_result(None)

        return message

    @_common_chan_call
    async def __call__(self, message):
        if self._closing:
            raise ClosedChannel
        if self.full:
            await self._input

        self._buffer.append(message)

        with suppress(EmptyWaiter):
            self._output(None)

    async def close(self):
        if self.closed or self._closing:
            return
        if len(self._buffer) > 0:
            self._closing = self.loop.create_future()
            await self._closing
        super().close_nowait()

    def __len__(self):
        return len(self._buffer)

    @property
    def full(self):
        return len(self._buffer) == self.maxsize

    @property
    def empty(self):
        return len(self._buffer) == 0
