from functools import wraps, partial
from collections import defaultdict
from inspect import isgeneratorfunction, iscoroutinefunction
import logging
from asyncio import get_event_loop, gather
from . import asyncify


class default_error_logger:
    """Default error handler which logs at DEBUG level."""

    __slots__ = "logger"
    kwargs = False
    once = False

    def __init__(self):
        self.logger = logging.getLogger("sweetasync.eventemitter")

    async def __call__(self, event):
        error, listener = event
        self.logger.debug("Exception in %s." % listener,
                          exc_info=error, stack_info=True)


class EventEmitter:
    """Event emitter inspired by the same class in node.js.

    Handling of backpressure is encouraged by :meth:`emit`, which
    will block until all listeners have returned. All exceptions
    are sent to listeners of the "error" event; its default listener
    logs the exception.

    :param loop: Event loop to use.
    :type loop: AbstractEventLoop or None
    :param bool log_errors: Set to ``False`` if you don't want
        the default error listener to be active. Note that you'll
        have to provide at least one for :meth:`emit` to work.

    This class should be inheritable without much difficulty as the
    public interface is as small as possible, and there are
    no private members and no complex metaprogramming.
    """

    __slots__ = "loop", "listeners"

    def __init__(self, log_errors=True, loop=None):
        self.loop = loop or get_event_loop()
        """Event loop used by the event emitter."""

        self.listeners = defaultdict(list)
        """
        Dict mapping event identifiers to lists of wrapped listeners.
        You can manipulate it to delete listeners or modify their priority.
        """

        if log_errors:
            logger = default_error_logger()
            # Don't use #on so as to avoid infinite recursion
            # if there's a bug somewhere.
            self.listeners["error"].append(logger)

    def on(self, event, listener=None, *, kwargs=False):
        """Add a listener for an event.

        Contrarily to node.js' EventEmitter, this method can't be chained.

        :param event: An event identifier.
        :type event: Any hashable
        :param listener: Callable that will be passed a message each time
            an event is triggered. It can be an asynchronous function.
            If not provided, this method will return a decorator which will
            add the decorated function as a listener.
        :param kwargs: Indicates if the listener can accept keyword arguments
            (this can't be reliably auto-detected).
        :type kwargs: bool
        :return: A decorated listener which launches the actual listener
            asynchronously and drops keyword arguments if they aren't accepted.
            You can use it to identify the listener in :attr:`listeners`.
        :raise AssertionError: The listener isn't a callable.
        """
        if listener is None:
            return partial(self.on, event, kwargs=kwargs)

        assert callable(listener)

        # Assume that generators are old-style coroutines, because else why
        # would you pass a generator to this method?
        if not (iscoroutinefunction(listener) or
                isgeneratorfunction(listener)):
            # We could be using loop.call_soon instead of adding yet another
            # wrapper, but we want :meth:`emit` to be able to return tasks.
            listener = asyncify(listener)

        @wraps(listener)
        async def wrapper(message, **kwargs):
            try:
                await listener(message, **kwargs)
            except BaseException as error:
                await self.emit("error", (error, wrapper))

        wrapper.kwargs = kwargs
        wrapper.once = False

        self.listeners[event].append(wrapper)

        return wrapper

    def once(self, event, listener=None, *, kwargs=False):
        """Add a listener for only one activation of an event.

        Otherwise works identically to :meth:`on`."""
        # Trying to refactor that as a decorator resulted in more
        # lines of code and more complexity.
        if listener is None:
            return partial(self.once, event, kwargs=kwargs)

        wrapper = self.on(event, listener, kwargs=kwargs)
        wrapper.once = True
        return wrapper

    async def emit(self, event, message, **kwargs):
        """Trigger an event, sending the same message to every listener.

        This method *must* be awaited, as it makes far easier to design
        systems which correctly propagate constraints in processing speed
        (ie backpressure). If you really want to use it in a "fire and forget"
        fashion, use :func:`asyncio.ensure_future`.

        Exceptions and their associated listener are sent in a tuple to
        listeners of "error".

        :param event: An event identifier.
        :type event: Any hashable
        :param message: An object that will be passed to every listener
            registered for that event. There is no guarantee the listeners
            will be launched in the order they were defined. Even if it does,
            don't rely on it.
        :param kwargs: Optionnal arguments for listeners; for those which don't
            accept keyword arguments, they won't raise an exception.
        :type kwargs: Anything that the listeners can accept
        :return: ``None``.
        :raises: :class:`AssertionError` if there is no listener for "error".
        """
        assert self.listeners["error"], "Error events must have a listener"
        tasks = list()

        # Make the list of listeners immutable else
        # deleting listeners for :meth:`once` will mess with the loop.
        for listener in tuple(self.listeners[event]):

            if listener.kwargs:
                task = listener(message, **kwargs)
            else:
                task = listener(message)

            tasks.append(task)

            if listener.once:
                self.listeners[event].remove(listener)

        await gather(*tasks, loop=self.loop)
