"""Bridges between the async and the sync worlds."""
from functools import wraps
from inspect import isawaitable
from collections.abc import Iterable
from contextlib import suppress
from asyncio import (
        get_event_loop, new_event_loop, wait_for,
        sleep, gather, Task, CancelledError
    )


def asyncify(callback):
    """A decorator to turn sync functions into async functions.

    Useful to avoid duplicating code or to create tasks.
    """
    @wraps(callback)
    async def wrapped(*args, **kwargs):
        return callback(*args, **kwargs)
    return wrapped


class aiorunner:
    """A decorator and context manager to turn async flow synchronous.

    :param timeout: Seconds before dropping the execution.
    :type timeout: int or None
    :param loop: Event loop to use.
    :type loop: AbstractEventLoop or None
    :param bool new_loop: If set will create a new loop on every run.
    :returns: A decorator doubling as a context manager.
    :raises: Any exception the coroutine will raise.

    When used as a context manager, you can block on awaitables by calling
    :meth:`run`. When used as a decorator, it turns async functions sync.
    """

    __slots__ = "timeout", "new_loop", "loop", "_prev_tasks"

    def __init__(self, timeout=None, loop=None, new_loop=False):
        if new_loop:
            self.loop = None
        else:
            self.loop = loop or get_event_loop()
        self.new_loop = new_loop
        """Boolean indicating if a new loop is created on each run."""
        self.timeout = timeout
        """Seconds before dropping the execution."""

    def _get_loop(self):
        if self.new_loop:
            self.loop = new_event_loop()
        return self.loop

    def _single_schedule(self, awaitable, timeout=False):
        timeout = self.timeout if timeout is False else timeout
        assert isawaitable(awaitable)
        return wait_for(awaitable, timeout, loop=self.loop)

    def schedule(self, awaitable, *awaitables, timeout=False):
        """Schedule one or several awaitables in the event loop.

        :param awaitables: Iterable of awaitables to schedule.
        :param int timeout: Timeout on the task(s). Defaults to the
            object's timeout.
        :return: A task gathering the awaitables.
        :raise: AssertionError if one argument isn't an awaitable.
        """
        if awaitables:
            tasks = [self._single_schedule(awtbl, timeout=timeout)
                     for awtbl in (awaitable, *awaitables)]
            return gather(*tasks, loop=self.loop)
        else:
            return self._single_schedule(awaitable, timeout=timeout)

    def run(self, task, *tasks, timeout=False):
        """Run the task with the choosen timeout and event loop.

        :param task: One or several awaitables to block on.
        :param type: Awaitables
        :param int timeout: Timeout on the task(s). Defaults to the
            object's timeout.
        :return: The result of the awaitable or a list of results.
        :raise: AssertionError if one argument isn't an awaitable.
        """
        task = self.schedule(task, *tasks, timeout=timeout)
        return self.loop.run_until_complete(task)

    def cancel(self):
        """Cancel all tasks added while in the context up to that point.

        :raise: Any exception but CancelledError.
        """
        remaining_tasks = Task.all_tasks(self.loop) - self._prev_tasks
        to_cancel = [task for task in remaining_tasks if not task.done()]
        if to_cancel:
            for task in to_cancel:
                task.cancel()
            with suppress(CancelledError):
                self.run(*to_cancel)

    def __enter__(self):
        self._get_loop()
        self._prev_tasks = Task.all_tasks(self.loop)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.cancel()
        finally:
            if self.new_loop:
                self.loop.close()
        return True if exc_type is None else False

    def __call__(self, aiofunc):
        @wraps(aiofunc)
        def syncified(*args, **kwargs):
            with self:
                return self.run(aiofunc(*args, **kwargs))
        return syncified


def sync(aiofunc):
    """Use :class:`aiorunner` with its default parameters.

    :param aiofunc: Coroutine function to be decorated.
    :return: A synchronous function.
    :raises: Any exception the coroutine will raise.
    """
    actx = aiorunner()
    wrapper = wraps(aiofunc)
    return wrapper(actx(aiofunc))


class aiter:
    """Turn an iterable into an asynchronous iterable.

    This is meant to spread intensive computation along the event loop.

    :param Iterable iterable: Any iterable.
    :return: An asynchronous iterator.
    :raises TypeError: You must provide an iterable.
    :raises StopAsyncIteration:

    Example::

        async for _ in aiter(range(10000)):
            do_something()
    """

    slots = "iterable"

    def __init__(self, iterable):
        if not isinstance(iterable, Iterable):
            raise TypeError("You must provide an iterable.")
        self.iterable = iter(iterable)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            await sleep(0)
            return next(self.iterable)
        except StopIteration:
            raise StopAsyncIteration


def as_await(method):
    """Decorator to use an async function to define an __await__ method."""
    @wraps(method)
    def wrapped(self):
        return method(self).__await__()
    return wrapped
