from asyncio import get_event_loop
from collections import deque


class EmptyWaiter(Exception):
    pass


class FullWaiter(Exception):
    pass


class Waiter:
    """An awaitable which accepts multiple results.

    :param loop: The event loop to use.
    :type loop: An event loop or ``None`` to use the default one
    :param maxsize: Maximum number of simultaneous listeners.
    :type maxsize: int or ``None`` for no limit

    The length of a waiter is the number of asynchronous listeners awaiting it.
    This class is subscriptable, and returns a future that was enqueued.
    Trying to send a value or get an item when there is no listener
    (or when no future has been enqueued with :meth:`get`) will raise
    :class:`EmptyWaiter`.

    Waiters have an asynchronous side and a synchronous side.
    It can be awaited multiple times, and each time the waiter is called on a
    value, this value is sent to the first non-cancelled listener.
    The values are sent synchronously.
    Typically used to have multiple coroutines wait in line to get a value.

    This class is useful to implement asynchronous primitives such as locks
    or channels and isn't meant to be used casually as you have to carefully
    coordinate the moment values are sent and manage the exceptions raised
    when the awaiting side isn't ready.

    Here is an example of an implementation simplified thanks to this class,
    a basic asynchronous lock::

        from sweetasync import Waiter, EmptyWaiter
        from contextlib import suppress

        class Lock:

            def __init__(self):
                self.active = False
                self._waiter = Waiter()

            async def acquire(self):
                if self.active:
                    await self._waiter
                self.active = True

            def release(self):
                with suppress(EmptyWaiter):
                    self._waiter(None)
                self.active = False
    """

    slots = "loop", "_queue", "maxsize"

    def __init__(self, loop=None, maxsize=None):
        self.loop = loop or get_event_loop()
        """The event loop the waiter is attached to."""
        self._queue = deque()
        self.maxsize = maxsize

    def __call__(self, result):
        """Input."""
        while True:
            try:
                future = self._queue.popleft()
            except IndexError:
                raise EmptyWaiter
            if not future.done():
                break
        future.set_result(result)

    def __await__(self):
        """Output."""
        return self.get().__await__()

    def get(self):
        """Create and enqueue a new future into the waiter, then return it.

        :return: A future that was enqueued into the waiter.
        :raises: :class:`FullWaiter` if there are more than ``maxsize``
            enqueued futures.
        """
        if len(self) == self.maxsize:
            raise FullWaiter
        queue = self._queue
        future = self.loop.create_future()
        queue.append(future)
        return future

    def exception(self, exception):
        """Send an exception to all awaiters."""
        while self._queue:
            future = self._queue.popleft()
            if not future.done():
                future.set_exception(exception)

    def cancel(self):
        """Cancel all awaiters."""
        while self._queue:
            future = self._queue.popleft()
            if not future.done():
                future.cancel()

    def __len__(self):
        return len(self._queue)

    def __getitem__(self, i):
        try:
            return self._queue[i]
        except IndexError:
            raise EmptyWaiter
